use std::fmt::Debug;

use binrw::{BinRead, BinWrite};
use diff::Diff;

use crate::{
    field_of::FieldOf,
    types::{Field, ObjectType, Point, F64, U32},
};

use super::ObjectCore;

#[cfg_attr(feature = "default-debug", derive(Debug))]
#[derive(BinRead, BinWrite, Diff, PartialEq)]
#[diff(attr(
    #[derive(Debug, PartialEq)]
))]
pub struct Polygon {
    pub core: ObjectCore,
    #[brw(magic(10u32))] // Number of following fields in struct
    pub invert_shape: U32,
    pub corner_a: FieldOf<Point>,
    pub corner_b: FieldOf<Point>,
    pub offset_cx: F64,
    pub offset_cy: F64,
    pub offset_dx: F64,
    pub offset_dy: F64,
    pub edges: U32,
    _unknown_1: [Field; 2],
}

// Custom Debug implementation to only print known fields
#[cfg(not(feature = "default-debug"))]
impl Debug for Polygon {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Polygon")
            .field("core", &self.core)
            .field("invert_shape", &self.invert_shape)
            .field("corner_a", &self.corner_a)
            .field("corner_b", &self.corner_b)
            .field("offset_cx", &self.offset_cx)
            .field("offset_cy", &self.offset_cy)
            .field("offset_dx", &self.offset_dx)
            .field("offset_dy", &self.offset_dy)
            .field("edges", &self.edges)
            .finish()
    }
}

impl Default for Polygon {
    fn default() -> Self {
        Self {
            core: ObjectCore {
                obj_type: ObjectType::Polygon.into(),
                ..Default::default()
            },
            invert_shape: 1.into(),
            corner_a: Point { x: 0.0, y: 0.0 }.into(),
            corner_b: Point { x: 0.0, y: 0.0 }.into(),
            offset_cx: 0.0.into(),
            offset_cy: 0.0.into(),
            offset_dx: 0.0.into(),
            offset_dy: 0.0.into(),
            edges: 6.into(),
            _unknown_1: [
                vec![0, 0, 0, 0].into(), // 0_u32
                vec![
                    0, 0, 0, 0, 0, 0, 240, 63, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 240, 63, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 240, 63,
                ] // ???
                .into(),
            ],
        }
    }
}
